import sys
from paos import logger


def addLogFile(fname="paos.log"):
    """
    Adds a new log file for logging.

    Parameters
    ----------
    fname : str, optional
        The filename for the log file. Defaults to 'paos.log'.

    Returns
    -------
    None
    """
    logger.add(fname)


def setLogLevel(level="INFO"):
    """
    Configures the logging level for the logger.

    Parameters
    ----------
    level : str, optional
        The logging level to set. Defaults to 'INFO'. Possible values include
        'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', etc.

    Returns
    -------
    None
    """
    logger.configure(handlers=[{"sink": sys.stderr, "level": level}])


def disableLogging(name="paos"):
    """
    Disables the logger for the given module name.

    Parameters
    ----------
    name : str, optional
        The module name to disable logging for. Defaults to 'paos'.

    Returns
    -------
    None
    """
    logger.disable(name)


def enableLogging(name="paos"):
    """
    Enables the logger for the given module name.

    Parameters
    ----------
    name : str, optional
        The module name to enable logging for. Defaults to 'paos'.

    Returns
    -------
    None
    """
    logger.enable(name)



import configparser
import numpy as np
from shiny import ui
import faicons as fa

ICONS = {
    "ellipsis": fa.icon_svg("ellipsis"),
    "gear": fa.icon_svg("gear"),
    "open": fa.icon_svg("folder-open"),
    "save": fa.icon_svg("floppy-disk"),
    "close": fa.icon_svg("xmark"),
    "docs": fa.icon_svg("book-open-reader"),
    "about": fa.icon_svg("info"),
    "run": fa.icon_svg("play"),
    "refresh": fa.icon_svg("arrows-rotate"),
    "info": fa.icon_svg("circle-info"),
    "load": fa.icon_svg("cloud-arrow-up"),
    "download": fa.icon_svg("cloud-arrow-down"),
    "wizard": fa.icon_svg("hat-wizard"),
}

CARD_HEADER_CLASS = "d-flex justify-content-between align-items-center"

vline = ui.HTML(
    """
    <div style="border-right: 0.1px solid lightgrey; height: 100%;"></div>
    """
)

hline = ui.HTML(
    """
    <div style="border-bottom: 0.1px solid lightgrey; width: 100%;"></div>
    """
)


vspace = ui.tags.div(style="height: 15px;")


def menu_panel(id):
    return ui.nav_panel(
        ui.input_action_button(
            id,
            id.capitalize(),
            icon=ICONS[id],
            width="100%",
        )
    )


def fill_value(items, name, row, col):
    func = items[name]["f"]
    prefix = "" if "prefix" not in items[name] else items[name]["prefix"]
    theid = f"{prefix}{name}_{row}_{col}"
    if func in [ui.input_select]:
        return func(
            id=theid,
            label=items[name]["label"] if "label" in items[name] else "",
            choices=items[name]["choices"],
            selected=items[name]["selected"],
        )
    elif func in [ui.tags.input]:
        return func(
            id=theid,
            type="text",
            value=items[name]["value"],
            readonly=items[name]["readonly"],
            class_="form-control",
        )
    elif func in [ui.input_text]:
        return func(
            id=theid,
            label=items[name]["label"] if "label" in items[name] else "",
            value=items[name]["value"],
            placeholder=items[name]["placeholder"] if "placeholder" in items[name] else "",
        )
    elif func in [ui.input_checkbox]:
        return func(
            id=theid,
            label=items[name]["label"] if "label" in items[name] else "",
            value=items[name]["value"],
        )
    elif func in [ui.popover]:
        return (
            func(
                ICONS["gear"],
                ui.div(
                    *[
                        fill_value(items[name]["value"], key, row, col)
                        for key in items[name]["value"].keys()
                    ],
                ),
                title=items[name]["title"],
                placement="top",
            ),
        )
    elif func in [ui.p]:
        return func(
            f"{items[name]['value']}",
        )
    elif func in [ui.navset_card_pill]:
        return func(
            ui.nav_panel(
                items[name]["title"],
                ui.div(
                    *[
                        fill_value(items[name]["value"], key, row, col)
                        for key in items[name]["value"].keys()
                    ],
                ),
            ),
            selected=False,
        )
    elif func in [ui.accordion]:
        return func(
            ui.accordion_panel(
                items[name]["title"],
                ui.div(
                    *[
                        fill_value(items[name]["value"], key, row, col)
                        for key in items[name]["value"].keys()
                    ],
                ),
            ),
            open=False,
        )
    elif func in [ui.card]:
        return func(
            ui.div(
                items[name]["title"],
                ui.div(
                    *[
                        fill_value(items[name]["value"], key, row, col)
                        for key in items[name]["value"].keys()
                    ],
                ),
            ),
            open=False,
        )


def fill_header(items):
    if not items:
        return ui.div()
    key = min(np.array(list(items.keys())))
    item = items[key]
    return ui.div(
        {"style": "display: flex;"},
        *[
            ui.column(
                subitem["width"],
                {"style": "text-align: center;"},
                ui.markdown(f"**{subkey}**"),
                hline,
            )
            for _, (subkey, subitem) in enumerate(item.items())
        ],
    )


def fill_body(items):
    if not items:
        return [ui.div()]
    return [
        ui.div(
            {"style": "display: flex;"},
            *[
                ui.column(
                    subitem["width"],
                    {"style": "text-align: center;"},
                    fill_value(item, subkey, row + 1, col),
                )
                for col, (subkey, subitem) in enumerate(item.items())
            ],
        )
        for row, (_, item) in enumerate(items.items())
    ]


def refresh_ui(name, items, mode=None, key=""):
    ui.remove_ui(f"#inserted-{name}-editor")

    func = ui.div

    if mode == "dict":
        items = [fill_header(items), *fill_body(items)]
        func = ui.card

    elif mode == "nested-dict":
        key = list(items.keys())[0] if key == "" else key
        items = [fill_header(items[key]), *fill_body(items[key])]
        func = ui.card

    elif mode == "body":
        items = [*fill_body(items)]

    elif mode == "header":
        items = [*fill_body(items)]

    ui.insert_ui(
        func(
            {"id": f"inserted-{name}-editor"},
            *items,
        ),
        selector=f"#{name}-editor",
        where="beforeEnd",
    )


def to_configparser(dictionary):
    """
    Given a dictionary, it converts it into a :class:`~configparser.ConfigParser` object

    Parameters
    ----------
    dictionary: dict
        input dictionary to be converted

    Returns
    -------
    out: :class:`~configparser.ConfigParser`
    """

    config = configparser.ConfigParser()

    for key, item in dictionary.items():
        if isinstance(item, dict):
            config.add_section(key)
            for subkey, subitem in item.items():
                if subitem is not None and subitem != "":
                    if isinstance(subitem, str):
                        pass
                    elif isinstance(subitem, (float, bool)):
                        subitem = str(subitem)
                    elif isinstance(subitem, (tuple, list)):
                        subitem = ",".join(subitem)
                    else:
                        raise NotImplementedError("item type not supported")
                config.set(key, subkey, subitem)

    return config


def ellipsis(id, names, choices):
    return ui.popover(
        ICONS["ellipsis"],
        *[
            ui.input_select(
                id=f"{id}_select_{name}",
                label=f"Select {name.capitalize()}",
                choices=choice,
            )
            for name, choice in zip(names, choices)
        ],
        title="",
        placement="top",
    )


def output_text_verbatim(id, placeholder=True):
    return ui.div(
        {"id": f"{id}-editor"},
        ui.div(
            {"id": f"inserted-{id}-editor"},
            ui.output_text_verbatim(
                id,
                placeholder=placeholder,
            ),
        ),
    )


def modal_download(id, ext):
    m = ui.modal(
        *[
            ui.input_text(
                id=f"save_{ext}",
                label="Save As",
                value=f"filename.{ext}",
                placeholder=f"filename.{ext}",
            ),
            ui.download_button(
                f"download_{id}_{ext}",
                "Download",
                icon=ICONS["download"],
                class_="btn-success",
            ),
            # ui.output_text_verbatim(f"download_{ext}_progress"),
        ],
        title=f"Save to {ext.upper()} File",
        easy_close=True,
    )
    ui.modal_show(m)


def nested_div(id):
    return ui.div(
        {"id": f"{id}-editor"},
        ui.div(
            {"id": f"inserted-{id}-editor"},
        ),
    )



from shiny import ui

from paos import Zernike
from paos import PSD

from paos.gui.core.shared import output_text_verbatim
from paos.gui.core.shared import nested_div
from paos.gui.core.shared import CARD_HEADER_CLASS
from paos.gui.core.shared import ICONS
from paos.gui.core.shared import vspace


Placeholder = {
    "Par1": {
        "INIT": "",
        "Coordinate Break": "Xdecenter",
        "Standard": "",
        "Paraxial Lens": "Focal length",
        "ABCD": "Ax",
        "Zernike": "Wavelength",
        "PSD": "A",
        "Grid Sag": "Wavelength",
    },
    "Par2": {
        "INIT": "",
        "Coordinate Break": "Ydecenter",
        "Standard": "",
        "Paraxial Lens": "",
        "ABCD": "Bx",
        "Zernike": "Ordering",
        "PSD": "B",
        "Grid Sag": "Nx",
    },
    "Par3": {
        "INIT": "",
        "Coordinate Break": "Xtilt",
        "Standard": "",
        "Paraxial Lens": "",
        "ABCD": "Cx",
        "Zernike": "Normalization",
        "PSD": "C",
        "Grid Sag": "Ny",
    },
    "Par4": {
        "INIT": "",
        "Coordinate Break": "Ytilt",
        "Standard": "",
        "Paraxial Lens": "",
        "ABCD": "Dx",
        "Zernike": "Radius of S.A.",
        "PSD": "fknee",
        "Grid Sag": "Dx",
    },
    "Par5": {
        "INIT": "",
        "Coordinate Break": "",
        "Standard": "",
        "Paraxial Lens": "",
        "ABCD": "Ay",
        "Zernike": "Origin",
        "PSD": "fmin",
        "Grid Sag": "Dy",
    },
    "Par6": {
        "INIT": "",
        "Coordinate Break": "",
        "Standard": "",
        "Paraxial Lens": "",
        "ABCD": "By",
        "Zernike": "Orthonorm",
        "PSD": "fmax",
        "Grid Sag": "Xdecenter (pix)",
    },
    "Par7": {
        "INIT": "",
        "Coordinate Break": "",
        "Standard": "",
        "Paraxial Lens": "",
        "ABCD": "Cy",
        "Zernike": "",
        "PSD": "SR",
        "Grid Sag": "Ydecenter (pix)",
    },
    "Par8": {
        "INIT": "",
        "Coordinate Break": "",
        "Standard": "",
        "Paraxial Lens": "",
        "ABCD": "Dy",
        "Zernike": "",
        "PSD": "units",
        "Grid Sag": "Errormap filepath",
    },
}


def app_elems(config):
    general_elems = [
        ui.input_text(
            "project",
            "Project name",
            value=config["general"].get("project", "Template"),
        ),
        ui.input_text("comment", "Comment", value=config["general"].get("comment", "")),
        ui.input_text(
            "version",
            "Version",
            value=config["general"].get("version", "0.1"),
        ),
        ui.input_text(
            "grid_size",
            "Grid size",
            value=config["general"].getint("grid_size", "512"),
        ),
        ui.input_text("zoom", "Zoom", value=config["general"].getint("zoom", "4")),
        ui.input_text(
            "tambient",
            "Ambient temperature [°C]",
            value=config["general"].getfloat("tambient", "20.0"),
        ),
        ui.input_text(
            "pambient",
            "Ambient pressure [atm]",
            value=config["general"].getfloat("pambient", "1.0"),
        ),
        ui.tags.div(
            ui.tags.label("Wavelength unit", class_="form-label"),
            ui.tags.input(
                id="wavelength_unit",
                type="text",
                value=config["general"].get("wavelength_unit", "micron"),
                readonly=True,
                class_="form-control",
            ),
        ),
        vspace,
        ui.tags.div(
            ui.tags.label("Angle unit", class_="form-label"),
            ui.tags.input(
                id="angle_unit",
                type="text",
                value=config["general"].get("angle_unit", "deg"),
                readonly=True,
                class_="form-control",
            ),
        ),
        vspace,
        ui.tags.div(
            ui.tags.label("Lens unit", class_="form-label"),
            ui.tags.input(
                id="lens_unit",
                type="text",
                value=config["general"].get("lens_unit", "m"),
                readonly=True,
                class_="form-control",
            ),
        ),
    ]

    field_elems = {}
    for n, (key, value) in enumerate(config.items("fields")):
        n += 1
        field_elems[n] = {}
        field_elems[n]["#"] = {
            "f": ui.p,
            "width": 5,
            "value": n,
        }
        field_elems[n]["Field"] = {
            "f": ui.input_text,
            "id": key,
            "value": value,
            "width": 7,
        }

    wl_elems = {}
    for n, (key, value) in enumerate(config.items("wavelengths")):
        n += 1
        wl_elems[n] = {}
        wl_elems[n]["#"] = {
            "f": ui.p,
            "width": 5,
            "value": n,
        }
        wl_elems[n]["Wavelength"] = {
            "f": ui.input_text,
            "id": key,
            "value": value,
            "width": 7,
        }

    lens_elems = {}
    for key in config.sections():
        if not key.startswith("lens_"):
            continue
        n = int(key.split("_")[1])
        item = config[key]

        lens_elems[n] = {}
        lens_elems[n]["#"] = {
            "f": ui.p,
            "width": 1,
            "value": n,
        }
        surface_type = item.get("surfacetype")
        lens_elems[n]["SurfaceType"] = {
            "f": ui.tags.input,
            # "choices": [
            #     "INIT",
            #     "Coordinate Break",
            #     "Standard",
            #     "Paraxial Lens",
            #     "ABCD",
            #     "Zernike",
            #     "PSD",
            #     "Grid Sag",
            # ],
            "width": 1,
            "value": surface_type,
            # "selected": surface_type,
            "prefix": "lens_",
            "readonly": True,
        }
        lens_elems[n]["Comment"] = {
            "f": ui.input_text,
            "value": item.get("comment"),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Radius"] = {
            "f": ui.input_text,
            "value": item.get("radius", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Thickness"] = {
            "f": ui.input_text,
            "value": item.get("thickness", ""),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Material"] = {
            "f": ui.input_select,
            "choices": [
                "",
                "MIRROR",
                "BK7",
                "BAF2",
                "CAF2",
                "SAPPHIRE",
                "SF11",
                "ZNSE",
            ],
            "width": 1,
            "selected": item.get("material", ""),
            "prefix": "lens_",
        }
        lens_elems[n]["Save"] = {
            "f": ui.input_checkbox,
            "value": item.getboolean("save", True),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Ignore"] = {
            "f": ui.input_checkbox,
            "value": item.getboolean("ignore", False),
            "width": 1,
            "prefix": "lens_",
        }
        lens_elems[n]["Stop"] = {
            "f": ui.input_checkbox,
            "value": item.getboolean("stop", False),
            "width": 1,
            "prefix": "lens_",
        }
        aperture = item.get("aperture", "")
        if aperture != "":
            aperture = " ".join(aperture.split())
            aperture = aperture.split(",")
            aperture_type = str(aperture[0])
            xhw = float(aperture[1])
            yhw = float(aperture[2])
            xdecenter = float(aperture[3])
            ydecenter = float(aperture[4])
        else:
            aperture_type = xhw = yhw = xdecenter = ydecenter = ""
        lens_elems[n]["Aperture"] = {
            "title": "Settings",
            "f": ui.popover,
            "width": 1,
            "value": {
                "Aperture_Type": {
                    "f": ui.input_select,
                    "selected": aperture_type,
                    "choices": [
                        "",
                        "rectangular aperture",
                        "rectangular obscuration",
                        "elliptical aperture",
                        "elliptical obscuration",
                    ],
                    "label": "Type",
                    "prefix": "lens_",
                },
                "Aperture_xhw": {
                    "f": ui.input_text,
                    "value": xhw,
                    "label": "X-Half Width",
                    "prefix": "lens_",
                },
                "Aperture_yhw": {
                    "f": ui.input_text,
                    "value": yhw,
                    "label": "Y-Half Width",
                    "prefix": "lens_",
                },
                "Aperture_xdecenter": {
                    "f": ui.input_text,
                    "value": xdecenter,
                    "label": "X-Decenter",
                    "prefix": "lens_",
                },
                "Aperture_ydecenter": {
                    "f": ui.input_text,
                    "value": ydecenter,
                    "label": "Y-Decenter",
                    "prefix": "lens_",
                },
            },
        }
        lens_elems[n]["Par1"] = {
            "f": ui.input_text,
            "value": item.get("par1", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par1"][surface_type],
        }
        lens_elems[n]["Par2"] = {
            "f": ui.input_text,
            "value": item.get("par2", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par2"][surface_type],
        }
        lens_elems[n]["Par3"] = {
            "f": ui.input_text,
            "value": item.get("par3", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par3"][surface_type],
        }
        lens_elems[n]["Par4"] = {
            "f": ui.input_text,
            "value": item.get("par4", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par4"][surface_type],
        }
        lens_elems[n]["Par5"] = {
            "f": ui.input_text,
            "value": item.get("par5", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par5"][surface_type],
        }
        lens_elems[n]["Par6"] = {
            "f": ui.input_text,
            "value": item.get("par6", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par6"][surface_type],
        }
        lens_elems[n]["Par7"] = {
            "f": ui.input_text,
            "value": item.get("par7", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par7"][surface_type],
        }
        lens_elems[n]["Par8"] = {
            "f": ui.input_text,
            "value": item.get("par8", ""),
            "width": 1,
            "prefix": "lens_",
            "placeholder": Placeholder["Par8"][surface_type],
        }

    zernike_elems = {}
    for key in config.sections():
        if not key.startswith("lens_"):
            continue
        item = config[key]
        if not item.get("surfacetype") == "Zernike":
            continue

        n = int(key.split("_")[1])
        zernike_elems[n] = {}

        zindex = item.get("zindex", "").split(",")
        zcoeff = item.get("z", "").split(",")

        ordering = item.get("par2")
        azimuthal, radial = Zernike.j2mn(N=len(zindex), ordering=ordering)

        for zi, zc in zip(zindex, zcoeff):
            zernike_elems[n][zi] = {}
            zernike_elems[n][zi]["n"] = {
                "f": ui.p,
                "width": 3,
                "value": radial[int(zi)],
            }
            zernike_elems[n][zi]["m"] = {
                "f": ui.p,
                "width": 3,
                "value": azimuthal[int(zi)],
            }
            zernike_elems[n][zi]["Zindex"] = {
                "f": ui.p,
                "width": 3,
                "value": zi,
            }
            zernike_elems[n][zi]["Zcoeff"] = {
                "f": ui.input_text,
                "width": 3,
                "value": zc,
                "prefix": f"lens_{n}_",
            }

    zernike_sidebar_elems = []
    zernike_tab_elems = []
    if zernike_elems:
        zernike_choices = [f"S{n}" for n in zernike_elems]

        zernike_sidebar_elems = [
            *[
                ui.input_select(
                    id=f"select_{name}", label=f"Select {name}", choices=choice
                )
                for name, choice in zip(["Zernike"], [zernike_choices])
            ],
        ]
        zernike_tab_elems = [
            ui.layout_column_wrap(
                ui.card(
                    ui.card_header("Explorer"),
                    ui.card_body(
                        output_text_verbatim("zernike_inputs"),
                        nested_div("zernike"),
                    ),
                    max_height="55vh",
                ),
                ui.card(
                    ui.card_header("Plot"),
                    ui.card_body(
                        output_text_verbatim("plot_zernike_inputs"),
                        ui.output_plot("plot_zernike"),
                    ),
                    ui.card_footer(
                        ui.input_action_button(
                            "do_plot_zernike",
                            "Run",
                            icon=ICONS["run"],
                            class_="btn-success",
                        ),
                        ui.input_action_button(
                            "download_plot_zernike",
                            "Save",
                            icon=ICONS["save"],
                        ),
                    ),
                    max_height="55vh",
                ),
            ),
        ]

    psd_elems = {}
    for key in config.sections():
        if not key.startswith("lens_"):
            continue
        item = config[key]
        if not item.get("surfacetype") == "PSD":
            continue

        n = int(key.split("_")[1])
        psd_elems[n] = {}
        psd_elems[n][0] = {}

        A = float(item.get("Par1"))
        B = float(item.get("Par2"))
        C = float(item.get("Par3"))
        fknee = float(item.get("Par4"))
        fmin = float(item.get("Par5"))
        fmax = float(item.get("Par6"))
        SR = item.get("Par7")
        units = item.get("Par8")
        sfe_rms = PSD.sfe_rms(A, B, C, fknee, fmin, fmax)

        # psd_elems[n][0]["A"] = {
        #     "f": ui.p,
        #     "width": 2,
        #     "value": f"{A:.3f}",
        # }
        # psd_elems[n][0]["B"] = {
        #     "f": ui.p,
        #     "width": 2,
        #     "value": f"{B:.3f}",
        # }
        # psd_elems[n][0]["C"] = {
        #     "f": ui.p,
        #     "width": 2,
        #     "value": f"{C:.3f}",
        # }
        # psd_elems[n][0]["fknee"] = {
        #     "f": ui.p,
        #     "width": 2,
        #     "value": f"{fknee:.3f}",
        # }
        # psd_elems[n][0]["fmin"] = {
        #     "f": ui.p,
        #     "width": 2,
        #     "value": f"{fmin:.3f}",
        # }
        # psd_elems[n][0]["fmax"] = {
        #     "f": ui.p,
        #     "width": 2,
        #     "value": f"{fmax:.3f}",
        # }
        psd_elems[n][0]["PSD SFE RMS"] = {
            "f": ui.p,
            "width": 2,
            "value": f"{sfe_rms:.3f}",
        }
        psd_elems[n][0]["SR RMS"] = {
            "f": ui.p,
            "width": 2,
            "value": f"{SR}",
        }
        psd_elems[n][0]["SFE units"] = {
            "f": ui.p,
            "width": 2,
            "value": f"{units}",
        }

    psd_sidebar_elems = []
    psd_tab_elems = []
    if psd_elems:
        psd_choices = [f"S{n}" for n in psd_elems]

        psd_sidebar_elems = [
            *[
                ui.input_select(
                    id=f"select_{name}", label=f"Select {name}", choices=choice
                )
                for name, choice in zip(["PSD"], [psd_choices])
            ],
        ]
        psd_tab_elems = [
            ui.layout_column_wrap(
                ui.card(
                    ui.card_header("Explorer"),
                    ui.card_body(
                        output_text_verbatim("psd_inputs"),
                        output_text_verbatim("psd_output"),
                        nested_div("PSD"),
                    ),
                    ui.card_footer(
                        ui.tags.div(
                            ui.input_action_button(
                                "calc_PSD",
                                "Run",
                                icon=ICONS["run"],
                                class_="btn-success",
                            ),
                        ),
                    ),
                    max_height="55vh",
                ),
                ui.card(
                    ui.card_header(
                        ui.tags.div(
                            "Plot",
                            ui.popover(
                                ICONS["gear"],
                                *[
                                    ui.input_text(
                                        id="psd_plot_phi",
                                        label="D [mm]",
                                        value=110.0,
                                    ),
                                ],
                                title="Settings",
                                placement="top",
                            ),
                            class_=CARD_HEADER_CLASS,
                        ),
                    ),
                    ui.card_body(
                        output_text_verbatim("plot_psd_inputs"),
                        ui.output_plot("plot_PSD"),
                    ),
                    ui.card_footer(
                        ui.input_action_button(
                            "do_plot_PSD",
                            "Run",
                            icon=ICONS["run"],
                            class_="btn-success",
                        ),
                        ui.input_action_button(
                            "download_plot_PSD", "Save", icon=ICONS["save"]
                        ),
                    ),
                    max_height="55vh",
                ),
            ),
        ]

    gridsag_elems = {}
    for key in config.sections():
        if not key.startswith("lens_"):
            continue
        item = config[key]
        if not item.get("surfacetype") == "Grid Sag":
            continue

        n = int(key.split("_")[1])
        gridsag_elems[n] = {}
        gridsag_elems[n][0] = {}

        filename = item.get("Par8")

        gridsag_elems[n][0]["Filename"] = {
            "f": ui.p,
            "width": 2,
            "value": f"{filename}",
        }

    gridsag_sidebar_elems = []
    gridsag_tab_elems = []
    if gridsag_elems:
        gridsag_choices = [f"S{n}" for n in gridsag_elems]

        gridsag_sidebar_elems = [
            *[
                ui.input_select(
                    id=f"select_{name}", label=f"Select {name}", choices=choice
                )
                for name, choice in zip(["gridsag"], [gridsag_choices])
            ],
        ]

        gridsag_tab_elems = [
            ui.layout_column_wrap(
                ui.card(
                    ui.card_header("Explorer"),
                    ui.card_body(
                        output_text_verbatim("gridsag_inputs"),
                        output_text_verbatim("gridsag_output"),
                        nested_div("gridsag"),
                    ),
                    max_height="55vh",
                ),
                ui.card(
                    ui.card_header(
                        "Plot",
                    ),
                    ui.card_body(
                        output_text_verbatim("plot_gridsag_inputs"),
                        ui.output_plot("plot_gridsag"),
                    ),
                    ui.card_footer(
                        ui.input_action_button(
                            "do_plot_gridsag",
                            "Run",
                            icon=ICONS["run"],
                            class_="btn-success",
                        ),
                        ui.input_action_button(
                            "download_plot_gridsag", "Save", icon=ICONS["save"]
                        ),
                    ),
                    max_height="55vh",
                ),
            ),
        ]

    wfe_elems = (
        zernike_sidebar_elems,
        zernike_elems,
        zernike_tab_elems,
        psd_sidebar_elems,
        psd_elems,
        psd_tab_elems,
        gridsag_sidebar_elems,
        gridsag_elems,
        gridsag_tab_elems,
    )

    surface_choices = [
        f"S{key}" for key, item in lens_elems.items() if item["Save"]["value"]
    ]

    wl_choices = [f"w{n}" for n in wl_elems]
    field_choices = [f"f{n}" for n in field_elems]

    analysis_sidebar_elems = [
        *[
            ui.input_select(id=f"select_{name}", label=f"Select {name}", choices=choice)
            for name, choice in zip(["field", "wl"], [field_choices, wl_choices])
        ],
    ]

    analysis_elems = [
        ui.layout_column_wrap(
            ui.navset_card_tab(
                ui.nav_panel(
                    "Fresnel POP",
                    ui.card(
                        ui.card_body(
                            output_text_verbatim("pop_inputs"),
                            output_text_verbatim("pop_output"),
                        ),
                        ui.card_footer(
                            ui.tags.div(
                                ui.input_action_button(
                                    "calc_pop",
                                    "Run",
                                    icon=ICONS["run"],
                                    class_="btn-success",
                                ),
                                ui.input_action_button(
                                    "download_pop", "Save", icon=ICONS["save"]
                                ),
                            ),
                        ),
                    ),
                ),
                ui.nav_panel(
                    "Ray Tracing",
                    ui.card(
                        ui.card_body(
                            output_text_verbatim("raytrace_inputs"),
                            output_text_verbatim("raytrace_output"),
                        ),
                        ui.card_footer(
                            ui.tags.div(
                                ui.input_action_button(
                                    "calc_raytrace",
                                    "Run",
                                    icon=ICONS["run"],
                                    class_="btn-success",
                                ),
                                ui.input_action_button(
                                    "download_raytrace", "Save", icon=ICONS["save"]
                                ),
                            ),
                        ),
                    ),
                ),
            ),
            ui.card(
                ui.card_header(
                    ui.tags.div(
                        "Plot",
                        ui.popover(
                            ICONS["gear"],
                            *[
                                ui.input_select(
                                    id="plot_select_surface",
                                    label="Select surface",
                                    choices=surface_choices,
                                    selected=surface_choices[-1],
                                ),
                                ui.input_select(
                                    id="plot_select_scale",
                                    label="Select scale",
                                    choices=["log", "linear"],
                                ),
                                ui.input_text(
                                    id="plot_select_zoom",
                                    label="Select zoom",
                                    value=1.0,
                                ),
                                ui.input_checkbox(
                                    id="plot_select_dark_rings",
                                    label="Dark rings",
                                    value=True,
                                ),
                            ],
                            title="Settings",
                            placement="top",
                        ),
                        class_=CARD_HEADER_CLASS,
                    ),
                ),
                ui.card_body(
                    output_text_verbatim("plot_inputs"),
                    ui.output_plot("plot"),
                ),
                ui.card_footer(
                    ui.tags.div(
                        ui.input_action_button(
                            "do_plot",
                            "Run",
                            icon=ICONS["run"],
                            class_="btn-success",
                        ),
                        ui.input_action_button(
                            "download_plot", "Save", icon=ICONS["save"]
                        ),
                    ),
                ),
            ),
        ),
    ]

    units_elems = []

    return (
        general_elems,
        field_elems,
        wl_elems,
        lens_elems,
        wfe_elems,
        analysis_sidebar_elems,
        analysis_elems,
        units_elems,
    )



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.patches import Ellipse
from astropy import units as u

from paos import logger
from paos import Zernike
from paos import PolyOrthoNorm
from paos import PSD


plt.rcParams["figure.facecolor"] = "white"
plt.rc("lines", linewidth=1.5)
plt.rc(
    "axes",
    axisbelow=True,
    titleweight="bold",
    labelcolor="dimgray",
    labelweight="bold",
)
plt.rc("font", size=12)
plt.rc("text", usetex=True)


def simple_plot(
    fig,
    axis,
    key,
    item,
    ima_scale,
    origin="lower",
    cmap="viridis",
    options={},
):
    """
    Given the POP simulation output dict, plots the squared amplitude of the
    wavefront at the given optical surface.

    Parameters
    ----------
    fig: :class:`~matplotlib.figure.Figure`
        instance of matplotlib figure artist
    axis: :class:`~matplotlib.axes.Axes`
        instance of matplotlib axes artist
    key: int
        optical surface index
    item: dict
        optical surface dict
    ima_scale: str
        plot color map scale, can be either 'linear' or 'log'
    origin: str
        matplotlib plot origin. Defaults to 'lower'
    cmap: str
        matplotlib plot color map. Defaults to 'viridis'
    options: dict
        dictionary containing the options to override the plotting default for one or more surfaces, specified by the
        dictionary key. Available options are the surface scale, an option to display physical units, the surface
        zoom(out), the plot scale and whether to plot dark rings in correspondance to the zeros of the Airy diffraction
        pattern.
        Examples:
        0) options={4: {'ima_scale':'linear'}}
        1) options={4: {'surface_scale':60, 'ima_scale':'linear'}}
        2) options={4: {'surface_scale':21, 'pixel_units':True, 'ima_scale':'linear'}}
        3) options={4: {'surface_zoom':2, 'ima_scale':'log', 'dark_rings': False}}

    Returns
    -------
    None
        updates the :class:`~matplotlib.figure.Figure` object

    Examples
    --------

    >>> from paos.core.parseConfig import parse_config
    >>> from paos.core.run import run
    >>> from paos.core.plot import simple_plot
    >>> pup_diameter, parameters, wavelengths, fields, opt_chains = parse_config('path/to/ini/file')
    >>> ret_val = run(pup_diameter, 1.0e-6 * wavelengths[0], parameters['grid size'],
    >>>              parameters['zoom'], fields[0], opt_chains[0])
    >>> from matplotlib import pyplot as plt
    >>> fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 8))
    >>> key = list(ret_val.keys())[-1]  # plot at last optical surface
    >>> item = ret_val[key]
    >>> simple_plot(fig, ax, key=key, item=item, ima_scale='log')
    >>> plt.show()

    """
    ap, wx, wy = None, None, None

    if key in options.keys() and "pixel_units" in options[key].keys():
        assert isinstance(
            options[key]["pixel_units"], bool
        ), "pixel_units must be boolean"
        pixel_units = options[key]["pixel_units"]
    else:
        pixel_units = False

    if key in options.keys() and "dark_rings" in options[key].keys():
        assert isinstance(
            options[key]["dark_rings"], bool
        ), "dark_rings must be boolean"
        dark_rings = options[key]["dark_rings"]
    else:
        dark_rings = True

    if item["wz"] < 0.005:
        # Use microns
        scale = 1.0e6
        unit = r"\textmu m"
    else:
        # Use mm
        scale = 1.0e3
        unit = "mm"

    if "psf" in item.keys():
        ima = np.ma.masked_array(data=item["psf"], mask=item["amplitude"] <= 0.0)
    else:
        ima = np.ma.masked_array(
            data=item["amplitude"] ** 2, mask=item["amplitude"] <= 0.0
        )
    power = ima.sum()

    if key in options.keys() and "ima_scale" in options[key].keys():
        assert isinstance(options[key]["ima_scale"], str), "ima_scale must be a str"
        assert options[key]["ima_scale"] in [
            "linear",
            "log",
        ], "ima_scale can be either linear or log"
        ima_scale = options[key]["ima_scale"]

    if ima_scale == "log":
        ima /= ima.max()
        im = axis.imshow(
            10 * np.ma.log10(ima),
            origin=origin,
            vmin=-20,
            vmax=0,
            cmap=plt.get_cmap(cmap),
        )
        cbar_label = "power/pix [db]"
    elif ima_scale == "linear":
        im = axis.imshow(ima, origin=origin, cmap=plt.get_cmap(cmap))
        cbar_label = "power/pix"
    else:
        logger.error("ima_scale shall be either log or linear")
        raise KeyError("ima_scale shall be either log or linear")

    fig.colorbar(im, ax=axis, orientation="vertical", label=cbar_label)

    if item["aperture"] is not None:
        x, y = item["aperture"].positions
        dx = 1.0 / scale if pixel_units else item["dx"]
        dy = 1.0 / scale if pixel_units else item["dy"]
        shapex = item["wfo"].shape[1]
        shapey = item["wfo"].shape[0]
        xy = scale * (x - shapex // 2) * dx, scale * (y - shapey // 2) * dy

        if hasattr(item["aperture"], "w") and hasattr(item["aperture"], "h"):
            w = item["aperture"].w
            h = item["aperture"].h
            wx = w * dx * scale
            wy = h * dy * scale
            xy = xy[0] - wx / 2, xy[1] - wy / 2
            ap = Rectangle(xy, wx, wy, ec="r", lw=5, fill=False)

        elif hasattr(item["aperture"], "a") and hasattr(item["aperture"], "b"):
            a = item["aperture"].a
            b = item["aperture"].b
            wx = 2 * a * dx * scale
            wy = 2 * b * dy * scale
            ap = Ellipse(xy, wx, wy, ec="r", lw=5, fill=False)

        axis.add_patch(ap)

    im.set_extent(np.array(item["extent"]) * scale)

    beam_radius = scale * item["wz"]
    airy_radius = 1.22 * scale * item["fratio"] * item["wl"]

    if np.isfinite(airy_radius) and dark_rings:
        for airy_scale in [1.22, 2.23, 3.24, 4.24, 5.24]:
            arad = airy_radius * airy_scale / 1.22
            width = 2.0 * arad / (scale * item["dx"]) if pixel_units else 2.0 * arad
            height = 2.0 * arad / (scale * item["dy"]) if pixel_units else 2.0 * arad
            aper = Ellipse(
                (0, 0),
                width=width,
                height=height,
                ec="k",
                lw=5,
                fill=False,
                alpha=0.5,
            )
            axis.add_patch(aper)

    if (
        beam_radius < airy_radius
        and (item["ABCDt"]() != np.eye(2)).all()
        and np.isfinite(airy_radius)
    ):
        plot_scale = 5.24 * airy_radius / 1.22  # 5th Airy null
    elif item["aperture"] is not None:
        plot_scale = max(wx / 2, wy / 2)
    else:
        plot_scale = beam_radius

    if key in options.keys():
        if "surface_zoom" in options[key].keys():
            plot_scale *= options[key]["surface_zoom"]
        elif "surface_scale" in options[key].keys():
            plot_scale = options[key]["surface_scale"]

    axis.set_title(
        rf"S{key:02d}"
        + "\n"
        + rf"F\#{item['fratio']:.2f} | w{scale * item['wz']:.2f}{unit:s} | "
        rf"$\lambda${1.0e6 * item['wl']:3.2f}\textmu m | P{100 * power:2.0f}\%",
        y=1.05,
    )

    if pixel_units:
        im.set_extent(
            np.array(item["extent"])
            / np.array([item["dx"], item["dx"], item["dy"], item["dy"]])
        )
        unit = "pixel"

    axis.set_xlabel(unit)
    axis.set_ylabel(unit)
    axis.set_xlim(-plot_scale, plot_scale)
    axis.set_ylim(-plot_scale, plot_scale)
    axis.grid()

    return


def zernike_plot(
    fig,
    axis,
    surface,
    index,
    Z,
    wavelength,
    ordering,
    normalize,
    orthonorm,
    grid_size,
):
    x = np.linspace(-1.0, 1.0, grid_size)
    xx, yy = np.meshgrid(x, x)
    rho = np.sqrt(xx**2 + yy**2)
    phi = np.arctan2(yy, xx)
    wavelength *= 1.0e-6  # convert to meters
    Z = np.array(Z) * wavelength

    func = PolyOrthoNorm if orthonorm else Zernike
    logger.debug(f"Using {func.__name__} polynomials")

    zernike = func(len(index), rho, phi, ordering=ordering, normalize=normalize)
    zer = zernike()
    wfe = (zer.T * Z).T.sum(axis=0)
    logger.debug(f"WFE RMS = {np.std(wfe)}")

    im = axis.imshow(wfe, origin="lower", cmap=plt.get_cmap("viridis"))
    fig.colorbar(im, ax=axis, orientation="vertical", label="WFE [m]")

    axis.set_title(
        f"{surface}" + "\n" + "Zernike errormap",
        y=1.05,
    )
    axis.set_xlabel("pixel")
    axis.set_ylabel("pixel")
    axis.grid()

    return


def psd_plot(
    fig,
    axis,
    surface,
    A,
    B,
    C,
    fknee,
    fmin,
    fmax,
    SR,
    units,
    grid_size,
    phi,
):
    phi_x = phi_y = phi  # mm
    D = np.max([phi_x, phi_y])
    delta = D / grid_size

    x = y = np.arange(-grid_size // 2, grid_size // 2) * delta
    xx, yy = np.meshgrid(x, y)
    pupil = np.zeros((grid_size, grid_size))

    mask = (2 * xx / phi_x) ** 2 + (2 * yy / phi_y) ** 2 <= 1

    pupil[mask] = 1.0
    wfo = np.ma.masked_array(pupil, mask=~mask)

    fx = np.fft.fftfreq(wfo.shape[0], delta)
    fxx, fyy = np.meshgrid(fx, fx)
    f = np.sqrt(fxx**2 + fyy**2)
    f[f == 0] = 1e-100

    wfe = PSD(wfo, A, B, C, f, fknee, fmin, fmax, SR, units=u.Unit(units))()

    im = axis.imshow(wfe, origin="lower", cmap=plt.get_cmap("viridis"))
    fig.colorbar(im, ax=axis, orientation="vertical", label="WFE [m]")

    axis.set_title(
        f"{surface}" + "\n" + "PSD errormap",
        y=1.05,
    )
    axis.set_xlabel("pixel")
    axis.set_ylabel("pixel")
    axis.grid()

    return


def gridsag_plot(
    fig,
    axis,
    surface,
    data,
):
    im = axis.imshow(data, origin="lower", cmap=plt.get_cmap("viridis"))
    fig.colorbar(im, ax=axis, orientation="vertical", label="Grid Sag [m]")

    axis.set_title(
        f"{surface}" + "\n" + "Grid Sag",
        y=1.05,
    )
    axis.set_xlabel("pixel")
    axis.set_ylabel("pixel")
    axis.grid()

    return
